<?php

namespace Drupal\message_digest\Plugin\Notifier;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\State\StateInterface;
use Drupal\message\MessageInterface;
use Drupal\message_digest\Exception\InvalidDigestGroupingException;
use Drupal\message_notify\Plugin\Notifier\MessageNotifierBase;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Message Digest notifier.
 */
abstract class DigestBase extends MessageNotifierBase implements ContainerFactoryPluginInterface, DigestInterface {

  /**
   * Database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The digest interval.
   *
   * @var string
   */
  protected $digestInterval;

  /**
   * The current request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The state service for tracking last sent time.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * Constructs the digest notifier plugins.
   *
   * @param array $configuration
   *   Plugin configuration array.
   * @param string $plugin_id
   *   The plugin ID.
   * @param mixed $plugin_definition
   *   The plugin definition.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   The message notify logger channel.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\message\MessageInterface $message
   *   (optional) The message entity.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, LoggerChannelInterface $logger, EntityTypeManagerInterface $entity_type_manager, MessageInterface $message = NULL, StateInterface $state, Connection $connection, RequestStack $request_stack) {
    // Set some defaults.
    $configuration += [
      'entity_type' => '',
      'entity_id' => '',
    ];

    parent::__construct($configuration, $plugin_id, $plugin_definition, $logger, $entity_type_manager, $message);
    $this->connection = $connection;
    $this->digestInterval = $plugin_definition['digest_interval'];
    $this->requestStack = $request_stack;
    $this->state = $state;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition, MessageInterface $message = NULL) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.channel.message_notify'),
      $container->get('entity_type.manager'),
      $message,
      $container->get('state'),
      $container->get('database'),
      $container->get('request_stack')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function deliver(array $output = []) {
    // Do not actually deliver this message because it will be delivered
    // via cron in a digest, but return TRUE to prevent a logged error.
    // Instead, we "deliver" it to the message_digest DB table so that it
    // can be retrieved at a later time.
    $message = $this->message;

    $message_digest = [
      'receiver' => $message->getOwnerId(),
      'entity_type' => $this->configuration['entity_type'],
      'entity_id' => $this->configuration['entity_id'],
      'notifier' => $this->getPluginId(),
      'timestamp' => $message->getCreatedTime(),
    ];

    // Don't allow entity_id without entity_type, or the reverse.
    if ($this->configuration['entity_type'] xor $this->configuration['entity_id']) {
      throw new InvalidDigestGroupingException(sprintf('Tried to create a message digest without both entity_type (%s) and entity_id (%s). These either both need to be empty, or have values.', $this->configuration['entity_type'], $this->configuration['entity_id']));
    }

    // Our $message is a cloned copy of the original $message with the mid field
    // removed to prevent overwriting (this happens in message_subscribe) so we
    // need to fetch the mid manually.
    $mid = $message->id();
    if (!$mid && isset($message->original)) {
      $mid = $message->original->id();
    }

    assert('!empty($mid)', 'The message entity (or $message->original) must be saved in order to create a digest entry.');
    $message_digest['mid'] = $mid;

    $this->connection->insert('message_digest')->fields($message_digest)->execute();

    return TRUE;
  }

  /**
   * The interval to compile digests for.
   *
   * @return string
   *   The interval. This should be compatible with `strtotime()`.
   */
  protected function getInterval() {
    return $this->digestInterval;
  }

  /**
   * {@inheritdoc}
   */
  public function aggregate() {
    if (!$this->processDigest()) {
      return [];
    }

    $interval = $this->getInterval();
    // Invert $interval since it's in the past.
    $start = strtotime('-' . $interval, $this->requestStack->getCurrentRequest()->server->get('REQUEST_TIME'));
    $message_groups = [];

    $query = $this->connection->select('message_digest', 'md');
    $query->fields('md');
    $query->condition('timestamp', $start, '>');
    $query->condition('sent', FALSE);
    $query->condition('notifier', $this->getPluginId());
    $result = $query->execute();

    foreach ($result as $row) {
      $entity_type = $row->entity_type;
      $entity_id = $row->entity_id;
      $account = $this->entityTypeManager->getStorage('user')->load($row->receiver);
      $context = [
        'data' => $row,
        // Set this to zero to aggregate group content.
        'entity_type' => $entity_type,
        'entity_id' => $entity_id,
      ];
      \Drupal::moduleHandler()->alter('message_digest_aggregate', $context, $account, $this);
      if (!empty($context['data']->mid)) {
        $message_groups[$context['data']->receiver][$context['entity_type']][$context['entity_id']][] = $context['data']->mid;
      }
    }
    return $message_groups;
  }

  /**
   * Determine if it is time to process this digest or not.
   *
   * @return bool
   *   Returns TRUE if a sufficient amount of time has passed.
   */
  protected function processDigest() {
    $interval = $this->getInterval();
    $key = $this->pluginId . '_last_run';
    $last_run = $this->state->get($key, 0);
    return $last_run < strtotime('-' . $interval, $this->requestStack->getCurrentRequest()->server->get('REQUEST_TIME'));
  }

  /**
   * {@inheritdoc}
   */
  public function markSent(UserInterface $account) {
    $this->connection->update('message_digest')
      ->fields(['sent' => TRUE])
      ->condition('receiver', $account->id())
      ->condition('notifier', $this->getPluginId())
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function setLastSent() {
    $this->state->set($this->getPluginId() . '_last_run', $this->requestStack->getCurrentRequest()->server->get('REQUEST_TIME'));
  }

}
